# -*- coding: utf-8 -*-
import datetime
import os
import logging
import time
from os.path import isfile, join
import shutil
import sys
from ftplib import FTP
from ftplib import FTP_TLS
from threading import Lock
from tornado.log import LogFormatter as _LogFormatter
from tempfile import TemporaryFile

from pyftpdlib.handlers import FTPHandler
from pyftpdlib.servers import FTPServer
from pyftpdlib.authorizers import DummyAuthorizer

from d22d.model import midhardware
from d22d.utils.utils import path_leaf, create_temporary_copy, time_stamp

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def set_file_log(log):
    # create a file handler
    handler = logging.FileHandler('ftp_log.log')
    handler.setLevel(logging.DEBUG)

    # create a logging format
    formatter = logging.Formatter('[%(asctime)s][%(name)s][%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)

    # add the handlers to the logger
    log.addHandler(handler)


class LogFormatter(_LogFormatter, object):
    def __init__(self, fmt=None, datefmt=None, color=True, *args, **kwargs):
        datefmt = "%Y-%m-%d %H:%M:%S"
        # '[%(threadName)s]' \
        fmt = '%(color)s[%(asctime)s][%(levelname)1.1s][%(name)s][%(filename)s:%(lineno)d:%(funcName)s()]%(end_color)s%(message)s'

        super(LogFormatter, self).__init__(color=color, fmt=fmt, datefmt=datefmt, *args, **kwargs)

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = str(datetime.datetime.now().strftime(datefmt))
        else:
            t = time.strftime("%Y-%m-%d %H:%M:%S", ct)
            s = "%s.%03d" % (t, record.msecs)
        return s


def set_shell_log(log):
    # create a file handler
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)

    # create a logging format
    handler.setFormatter(LogFormatter())

    # add the handlers to the logger
    log.addHandler(handler)
    log.propagate = False


def to_simple_chinese(string):
    # otherLanguage 变成中文
    item = string.encode('iso-8859-1').decode('gbk')
    return item


def to_latin_language(string):
    # 简体中文变成 latin
    item = string.encode('gbk').decode('iso-8859-1')
    return item


class FtpController:
    # /!\ Although the comments and variable names say 'file_name'/'file_anything' it inculdes folders also
    # Some functions in this class has no exception handling, it has to be done outside

    def __init__(self, host, port=21, username=' ', password=' ', use_tls=False, pasv=True, encoding='utf-8'):
        # List to store file search and search keywords
        self.search_file_list = []
        self.detailed_search_file_list = []
        self.keyword_list = []

        # Variable to hold the max no character name in file list (used for padding in GUIs)
        self.max_len = 0

        self.max_len_name = ''

        # Variable to tell weather hidden files are enabled
        self.hidden_files = False

        # Variable to store the platform the server is running on
        self.server_platform = 'Linux'
        self.lock = Lock()
        self.ftp = None
        self.remote_directory_path = '/'
        self.work_dir_now = '/'
        self.filesMoved = 0
        self.bytes_downloaded = 0
        self.bytes_uploaded = 0

        self.host = host
        self.username = username
        self.password = password
        self.port = port
        self.use_tls = use_tls
        self.pasv = pasv
        self.encoding = encoding

    def connect_to(self):
        try:
            # 先无tls，在试试有tls
            self.ftp = None
            if not self.use_tls:
                self.ftp = FTP(timeout=25)
                print(self.host, self.port)
                self.ftp.connect(self.host, self.port)
            if self.ftp is None:
                logger.info("LOG moveFTPFiles 3TLS ftp null")
                self.use_tls = True
                self.ftp = FTP_TLS(timeout=25)
                self.ftp.connect(self.host, self.port)
            self.ftp.login(self.username, self.password)

            if self.use_tls:
                self.ftp.prot_p()
            self.ftp.cwd(self.remote_directory_path)
            self.ftp.set_pasv(self.pasv)
            self.ftp.encoding = self.encoding
            logger.info(f"Connected. now pwd: {self.remote_directory_path}")

        except Exception as inst:
            logger.exception(f"Couldn't find server: {type(inst)}")  # the exception instance

    def try_all_connect_to(self):
        for tls, pasv in [
            (False, False),
            (False, True),
            (True, False),
            (True, True),
        ]:
            if self.try_connect_to(tls, pasv):
                print(f'尝试登录FTP成功：tls-{tls} pasv-{pasv}')
                return tls, pasv

    def try_connect_to(self, use_tls, pasv, encoding="utf-8"):
        try:
            # 先无tls，在试试有tls
            self.ftp = None
            if not use_tls:
                self.ftp = FTP(timeout=25)
                self.ftp.connect(self.host, self.port)
            else:
                self.use_tls = True
                self.ftp = FTP_TLS(timeout=25)
                self.ftp.connect(self.host, self.port)
            self.ftp.login(self.username, self.password)

            if use_tls:
                self.ftp.prot_p()
            self.ftp.cwd(self.remote_directory_path)
            self.ftp.set_pasv(pasv)
            self.ftp.encoding = encoding
            logger.info(f"Connected. now pwd: {self.remote_directory_path}")
            self.use_tls = use_tls
            self.pasv = pasv
            self.encoding = encoding
            return True
        except Exception as inst:
            logger.exception(f"Couldn't find server: {type(inst)}")  # the exception instance
            return False

    def check_encoding(self, encoding="utf-8"):
        try:
            self.ftp.encoding = encoding

            # TODO 新建中文文件夹

            for file in fs.client.get_detailed_file_list():
                print(file)
            self.encoding = encoding
        except Exception as inst:
            logger.exception(f"Couldn't find server: {type(inst)}")  # the exception instance

    def disconnect(self):
        self.ftp.quit()  # Close FTP connection
        self.ftp = None

    def toggle_hidden_files(self):
        self.hidden_files = not self.hidden_files

    def get_detailed_file_list(self, ignore_hidden_files_flag=False):
        files = []

        def dir_callback(line):
            if self.server_platform != 'Linux':
                files.append(line)
                return
            if (self.hidden_files is True or line.split()[8][0] != '.') or ignore_hidden_files_flag is True:
                files.append(line)

        self.ftp.dir(dir_callback)
        return files

    def get_file_list(self, detailed_file_list):
        self.max_len = 0
        self.max_len_name = ''
        file_list = []
        for x in detailed_file_list:
            # Remove details and append only the file name
            if self.server_platform == 'Linux':
                name = self.get_properties(x)[0]
            else:
                name = x
            file_list.append(name)
            if len(name) > self.max_len:
                self.max_len = len(name)
                self.max_len_name = name
        return file_list

    def get_detailed_search_file_list(self):
        return self.detailed_search_file_list

    def get_search_file_list(self):
        self.max_len = 0
        self.max_len_name = ''
        for name in self.search_file_list:
            if len(name) > self.max_len:
                self.max_len = len(name)
                self.max_len_name = name
        return self.search_file_list

    def chmod(self, filename, permissions):
        self.ftp.sendcmd('SITE CHMOD ' + str(permissions) + ' ' + filename)

    def is_there(self, path):
        try:
            self.ftp.sendcmd('MLST ' + path)
            return True
        except:
            return False

    def rename_dir(self, rename_from, rename_to):
        self.ftp.sendcmd('RNFR ' + rename_from)
        self.ftp.sendcmd('RNTO ' + rename_to)

    def move_dir(self, rename_from, rename_to, status_command, replace_command):
        if self.is_there(rename_to) is True:
            if replace_command(rename_from, 'File/Folder exists in destination folder') is True:
                self.delete_dir(rename_to, status_command)
            else:
                return
        try:
            self.ftp.sendcmd('RNFR ' + rename_from)
            self.ftp.sendcmd('RNTO ' + rename_to)
            status_command(rename_from, 'Moved')
        except:
            status_command(rename_from, 'Failed to move')

    def copy_file(self, file_dir, copy_from, file_size, status_command, replace_command):
        # Change to script's directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        if not os.path.exists('copy_temps'):
            os.makedirs('copy_temps')
        os.chdir('copy_temps')
        # Save the current path so that we can copy later
        dir_path_to_copy = self.ftp.pwd()
        # Change to the file's path and download it
        self.ftp.cwd(file_dir)
        self.download_file(copy_from, file_size, status_command, replace_command)
        # Change back to the saved path and upload it
        self.ftp.cwd(dir_path_to_copy)
        self.upload_file(copy_from, file_size, status_command, replace_command)
        # Delete the downloaded file
        os.remove(copy_from)
        status_command(copy_from, 'Deleted local file')

    def copy_dir(self, file_dir, copy_from, status_command, replace_command):
        # Change to script's directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        if not os.path.exists('copy_temps'):
            os.makedirs('copy_temps')
        os.chdir('copy_temps')
        # Save the current path so that we can copy later
        dir_path_to_copy = self.ftp.pwd()
        # Change to the file's path and download it
        self.ftp.cwd(file_dir)
        self.download_dir(copy_from, status_command, replace_command)
        # Change back to the saved path and upload it
        self.ftp.cwd(dir_path_to_copy)
        self.upload_dir(copy_from, status_command, replace_command)
        # Delete the downloaded folder
        shutil.rmtree(copy_from)
        status_command(copy_from, 'Deleting local directory')

    def delete_file(self, file_name, status_command):
        try:
            self.ftp.sendcmd('DELE ' + file_name)
            status_command(file_name, 'Deleted')
        except:
            status_command(file_name, 'Failed to delete')

    def delete_dir(self, dir_name, status_command):
        # Go into the directory
        self.ftp.cwd(dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list(True)
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            # If directory
            if self.is_dir(file_details):
                self.delete_dir(file_name, status_command)
            # If file
            else:
                self.delete_file(file_name, status_command)
        # Go back to parent directory and delete it
        try:
            self.ftp.cwd('..')
            status_command(dir_name, 'Deleting directory')
            self.ftp.sendcmd('RMD ' + dir_name)
        except:
            status_command(dir_name, 'Failed to delete directory')
            return

    def upload_file(self, file_name, file_size, status_command, replace_command):
        def update_progress(data):
            self.bytes_uploaded += int(sys.getsizeof(data))
            status_command(file_name, str(min(round((self.bytes_uploaded / file_size) * 100, 8), 100)) + '%')

        # Variable to keep trak of number of bytes uploaded
        self.bytes_uploaded = 0
        # Check if the file is already present in ftp server
        if self.is_there(file_name):
            if replace_command(file_name, 'File exists in destination folder') is False:
                return
        # Try to open file, if fails return
        try:
            file_to_up = open(file_name, 'rb')
        except:
            status_command(file_name, 'Failed to open file')
            return
        # Try to upload file
        try:
            status_command(file_name, 'Uploading')
            self.ftp.storbinary('STOR ' + file_name, file_to_up, 8192, update_progress)
            status_command(None, 'newline')
        except:
            status_command(file_name, 'Upload failed')
            return
        # Close file
        file_to_up.close()

    def upload_dir(self, dir_name, status_command, replace_command):
        # Change to directory
        os.chdir(dir_name)
        # Create directory in server and go inside
        try:
            if not self.is_there(dir_name):
                self.ftp.mkd(dir_name)
                status_command(dir_name, 'Creating directory')
            else:
                status_command(dir_name, 'Directory exists')
            self.ftp.cwd(dir_name)
        except:
            status_command(dir_name, 'Failed to create directory')
            return
        # Cycle through items
        for filename in os.listdir():
            # If file upload
            if isfile(filename):
                self.upload_file(filename, os.path.getsize(filename), status_command, replace_command)
            # If directory, recursive upload it
            else:
                self.upload_dir(filename, status_command, replace_command)

        # Got to parent directory
        self.ftp.cwd('..')
        os.chdir('..')

    def download_file(self, ftp_file_name, file_size, status_command, replace_command):
        # Function for updating status and writing to file
        def write_file(data):
            self.bytes_downloaded += int(sys.getsizeof(data))
            status_command(ftp_file_name, str(min(round((self.bytes_downloaded / file_size) * 100, 8), 100)) + '%')
            file_to_down.write(data)

        # Variable to keep track of total bytes downloaded
        self.bytes_downloaded = 0
        # Check if the file is already present in local directory
        if isfile(ftp_file_name):
            if replace_command(ftp_file_name, 'File exists in destination folder') is False:
                return
        # Try to open file, if fails return
        try:
            file_to_down = open(ftp_file_name, 'wb')
        except:
            status_command(ftp_file_name, 'Failed to create file')
            return
        # Try to upload file
        try:
            status_command(ftp_file_name, 'Downloading')
            self.ftp.retrbinary('RETR ' + ftp_file_name, write_file)
            status_command(None, 'newline')
        except:
            status_command(ftp_file_name, 'Download failed')
        # Close file
        file_to_down.close()

    def download_dir(self, ftp_dir_name, status_command, replace_command):
        # Create local directory
        try:
            if not os.path.isdir(ftp_dir_name):
                os.makedirs(ftp_dir_name)
                status_command(ftp_dir_name, 'Created local directory')
            else:
                status_command(ftp_dir_name, 'Local directory exists')
            os.chdir(ftp_dir_name)
        except:
            status_command(ftp_dir_name, 'Failed to create local directory')
            return
        # Go into the ftp directory
        self.ftp.cwd(ftp_dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list(True)
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            # If directory
            if self.is_dir(file_details):
                self.download_dir(file_name, status_command, replace_command)
            # If file
            else:
                self.download_file(file_name, int(self.get_properties(file_details)[3]), status_command,
                                   replace_command)
        # Got to parent directory
        self.ftp.cwd('..')
        os.chdir('..')

    def search(self, dir_name, status_command, search_file_name):
        # Go into the ftp directory
        self.ftp.cwd(dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list()
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            # If file_name matches the keyword, append it to search list
            if search_file_name.lower() in file_name.lower():
                if self.ftp.pwd() == '/':
                    dir = ''
                else:
                    dir = self.ftp.pwd()
                self.search_file_list.append(dir + '/' + file_name)
                self.detailed_search_file_list.append(file_details)
                status_command(dir + '/' + file_name, 'Found')
                # If directory, search it
            if self.is_dir(file_details):
                status_command(file_name, 'Searching directory')
                self.search(file_name, status_command, search_file_name)
        # Goto to parent directory
        self.ftp.cwd('..')

    def clear_search_list(self):
        del self.search_file_list[:]
        del self.detailed_search_file_list[:]

    def get_dir_size(self, dir_name):
        size = 0
        # Go into the ftp directory
        self.ftp.cwd(dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list()
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            if self.is_dir(file_details):
                size += self.get_dir_size(file_name)
            else:
                size += int(self.get_properties(file_details)[3])
        # Goto to parent directory
        self.ftp.cwd('..')
        # return size
        return size

    def cwd_parent(self, name):
        if '/' not in name:
            return name
        parent_name = '/'.join(name.split('/')[:-1])
        if parent_name == '':
            parent_name = '/'
        self.ftp.cwd(parent_name)
        return ''.join(name.split('/')[-1:])

    def mkd(self, name):
        self.ftp.mkd(name)

    def pwd(self):
        return self.ftp.pwd()

    def get_properties(self, file_details):
        if self.server_platform == 'Linux':
            details_list = file_details.split()
            # Get file attributes
            file_attribs = details_list[0]
            # Get date modified
            date_modified = ' '.join(details_list[5:8])
            # Remove the path from the name
            file_name = ' '.join(details_list[8:])
            # Get size if it is not a directory
            if 'd' not in file_details[0]:
                file_size = details_list[4]
                return [file_name, file_attribs, date_modified, file_size]
            else:
                return [file_name, file_attribs, date_modified]

    def is_dir(self, file_details):
        if self.server_platform == 'Linux':
            return 'd' in file_details[0]

    def cwd(self, path):
        try:
            self.ftp.cwd(path)
            self.work_dir_now = path
        except Exception as ex:
            logger.exception(ex)

    def moveFTPFiles(self, fileToUpload, remoteDirectoryPath):
        self.lock.acquire(True)
        try:
            localFileName = path_leaf(fileToUpload)
            if localFileName != "":
                # create a copy of the file before sending it
                # tempFile = create_temporary_copy(fileToUpload, workingDir, localFileName)
                # open a the local file
                with open(fileToUpload, 'rb') as rf:
                    wf = TemporaryFile('w+b')
                    while datar := rf.read(8128):
                        wf.write(datar)
                    wf.seek(0)
                    # Download the file a chunk at a time using RETR

                    # p_start = self.pwd()
                    p_start = self.work_dir_now
                    self.cwd(remoteDirectoryPath)
                    self.ftp.storbinary('STOR ' + localFileName, wf)
                    self.cwd(p_start)

                    wf.close()
                logger.info("Uploaded file: " + str(fileToUpload) + " on " + time_stamp())

                # Close the file
                # remove the temp file
                # to remove it I need to have the right permissions
                # os.chmod(tempFile, 777)
                # os.remove(tempFile)
        except Exception as inst:
            logger.error(type(inst))  # the exception instance
            logger.error(inst.args)  # arguments stored in .args
            logger.error(inst)  # __str__ allows args to be printed directly
            logger.error("Connection Error - " + time_stamp())

        self.lock.release()


class FtpClientStore(midhardware.BaseStore):
    def __init__(self, host, port, username, password, location='/', use_tls=False, pasv=True, encoding='utf-8'):
        self.client = FtpController(host, port, username, password, use_tls, pasv, encoding)
        self.location = location

    def count_data(self, data_type=None, *args, **kwargs):
        return NotImplementedError

    def list_data(self, data_type=None, *args, **kwargs):
        return NotImplementedError

    def get_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def save_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def delete_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def get_position(self, position, data_type=None, *args, **kwargs):
        return os.path.join(self.location, position)

    def get_data_size(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def check_self(self, *args, **kwargs):
        return NotImplementedError

    def save_self(self, *args, **kwargs):
        return NotImplementedError

    def free_self(self, *args, **kwargs):
        return NotImplementedError


class FtpSeverStore(midhardware.BaseStore):
    def __init__(
            self, host='0.0.0.0', port=21, username='user', password="123qwe", location='/home/user',
            use_tls=False, pasv=True, encoding='utf-8'):
        authorizer = DummyAuthorizer()
        authorizer.add_user(username, password, location, perm=password)
        handler = FTPHandler
        handler.authorizer = authorizer

        server = FTPServer((host, port), handler)
        server.serve_forever()
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.location = location

    def count_data(self, data_type=None, *args, **kwargs):
        return NotImplementedError

    def list_data(self, data_type=None, *args, **kwargs):
        return NotImplementedError

    def get_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def save_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def delete_data(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def get_position(self, position, data_type=None, *args, **kwargs):
        return os.path.join(self.location, position)

    def get_data_size(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def check_self(self, *args, **kwargs):
        return NotImplementedError

    def save_self(self, *args, **kwargs):
        return NotImplementedError

    def free_self(self, *args, **kwargs):
        return NotImplementedError


if __name__ == '__main__':
    set_shell_log(logger)

    __fs = FtpClientStore('127.0.0.1', 21111, 'python', '123456', '/', use_tls=False, pasv=True, encoding='utf-8')
    __fs.client.try_all_connect_to()
    for __f in __fs.client.get_detailed_file_list():
        print(__f)
